from __future__ import absolute_import, print_function, unicode_literals

import functools
import re

from requests import ConnectionError

from .exceptions import (
    ClientUnauthorized,
    DoesNotExist,
    GatewayDown,
    InvalidChannelName,
    InvalidDatabaseName,
    InvalidDocumentID,
    RevisionMismatch,
    SyncGatewayClientErrorResponse,
)


class ComparableMixin:
    """
    Alex Martelli's suggestion from https://stackoverflow.com/a/1061350/1286705
    """

    def __eq__(self, other):
        return not self < other and not other < self

    def __ne__(self, other):
        return self < other or other < self

    def __gt__(self, other):
        return other < self

    def __ge__(self, other):
        return not self < other

    def __le__(self, other):
        return not other < self


def sg_method(func, *args, **kwargs):
    """
    Wrap a normal request Verb (E.g. ``requests.get``) in exception handling
    logic.

    Raises:
        DoesNotExist: 404 was received for a request.
        GatewayDown: SyncGateway can't be reached.
        SyncGatewayClientErrorResponse: When any "not OK" response (according
            to ``requests.Response.ok`` is received that is not a 404.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            response = func(*args, **kwargs)
        except ConnectionError as ce:
            raise GatewayDown(ce.message)

        if response.status_code == 401:
            raise ClientUnauthorized(response.json()['reason'])
        elif response.status_code == 404:
            raise DoesNotExist('{} not found'.format(response.url))
        elif response.status_code == 409:
            raise RevisionMismatch()
        if not response.ok:
            raise SyncGatewayClientErrorResponse.from_response(response)
        return response

    return wrapper


def assert_valid_database_name(name):
    """
    Raise exception if database name does not match Couchbase requirements

    From docs (http://docs.couchdb.org/en/stable/api/database/common.html#put--db):

        The database name must begin with a lowercase letter.

        The database name must contain only valid characters. The following
        characters are valid in database names:

        * Lowercase letters: ``a-z``

        * Numbers: ``0-9``

        * Special characters: ``_$()+-/``

    NOTE hyphen is allowed, compared to channel names below.

    Args:
        name (str)

    Returns:
        ``None``

    Raises:
        InvalidDatabaseName: When passed ``name`` is invalid as a name for a
            database in Sync Gateway.

    Examples:

            >>> from pysyncgateway.helpers import assert_valid_database_name

        #.  Empty string is invalid:

            >>> assert_valid_database_name('')
            Traceback (most recent call last):
            ...
            InvalidDatabaseName: Empty database name is not allowed

        #.  Special-special characters are not allowed:

            >>> assert_valid_database_name('stuff!db')
            Traceback (most recent call last):
            ...
            InvalidDatabaseName: Special characters are not allowed in database names, first bad character is "!"

        #.  White space is not allowed:

            >>> assert_valid_database_name('db 2')
            Traceback (most recent call last):
            ...
            InvalidDatabaseName: Special characters are not allowed in database names, first bad character is " "

        #.  Capitals are not allowed:

            >>> assert_valid_database_name('bIGdATA')
            Traceback (most recent call last):
            ...
            InvalidDatabaseName: Special characters are not allowed in database names, first bad character is "I"

        #.  Name must start with a letter:

            >>> assert_valid_database_name('-10degrees')
            Traceback (most recent call last):
            ...
            InvalidDatabaseName: Database names must start with a lowercase letter

        #.  Happy 'simple' database names are OK:

            >>> assert_valid_database_name('construct-pm($)//x_x')
    """
    if not name:
        raise InvalidDatabaseName('Empty database name is not allowed')

    if not re.match(r'^[a-z]', name):
        raise InvalidDatabaseName('Database names must start with a lowercase letter')

    valid_db_name = re.match(r'^[a-z0-9\+\-\(\)\$/_]*', name).group()
    if valid_db_name == name:
        return

    raise InvalidDatabaseName(
        'Special characters are not allowed in database names, first bad character is "{}"'.format(
            name[len(valid_db_name)]
        )
    )


def assert_valid_document_id(doc_id):
    """
    NOTE this could become prohibitive if the admin client can't delete
    documents that fall outside the boundaries of the set of allowed names.
    E.g. App creates bad document and admin needs to clean it up, but can't.
    Therefore this validation should only be used on creation.

    Args:
        doc_id (str)

    Returns:
        ``None``

    Raises:
        InvalidDocumentID: When provided ``doc_id`` is not a valid ID for a
            Document in Sync Gateway.

    Examples:

            >>> from pysyncgateway.helpers import assert_valid_document_id

        #.  Empty string is invalid

            >>> assert_valid_document_id('')
            Traceback (most recent call last):
            ...
            InvalidDocumentID: Empty document id is not allowed

        #.  Colon is not allowed in document id

            >>> assert_valid_document_id('colon:nope')
            Traceback (most recent call last):
            ...
            InvalidDocumentID: Colon is not allowed in document ids

        #.  Question mark is not allowed (this is allowed in couchbase, but would
            require much urlencoding and mashing to get working via python IMO so
            banning it for now)

            If trying to quote, then ``quote(doc_id, safe='')`` will raise ``KeyError``
            and Hell's Armies walk the earth.

            >>> assert_valid_document_id('Questions? Nope')
            Traceback (most recent call last):
            ...
            InvalidDocumentID: Question mark is not allowed in document ids

        #.  Hash banned for the same reason as question mark.

            >>> assert_valid_document_id('hashes### like a boss')
            Traceback (most recent call last):
            ...
            InvalidDocumentID: Hash is not allowed in document ids

        #.  Special-special characters are allowed, and can start the doc id.

            >>> assert_valid_document_id('$stuff!channel')

        #.  White space is allowed.

            >>> assert_valid_document_id('channel 2')

        #.  Loads of special characters are OK

            >>> assert_valid_document_id('*-=|+/.,@(1234)')
    """
    if not doc_id:
        raise InvalidDocumentID('Empty document id is not allowed')

    for char, name in [(':', 'Colon'), ('?', 'Question mark'), ('#', 'Hash')]:
        if char in doc_id:
            raise InvalidDocumentID('{} is not allowed in document ids'.format(name))


def assert_valid_channel_name(name):
    """
    Assert channel name passes Sync Gateway's requirements:

        Valid channel names consist of letter ``[A-Z, a-z]``, digits ``[0-9]``,
        and a few special characters ``[= + / . , _ @]``. The empty string is
        not allowed. The special channel name ``*`` denotes all channels.

    Docs are out of date: https://github.com/couchbase/sync_gateway/issues/656
    Therefore adding ``-`` as allowed.

    Args:
        name (str)

    Raises:
        InvalidChannelName

    Examples:

            >>> from pysyncgateway.helpers import assert_valid_channel_name

        #.  Empty string is invalid:

            >>> assert_valid_channel_name('')  # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            InvalidChannelName: Empty channel name is not allowed

        #.  Special-special characters are not allowed:

            >>> assert_valid_channel_name('stuff!channel')
            Traceback (most recent call last):
            ...
            InvalidChannelName: Special characters are not allowed in channels, first bad character is "!"

        #.  White space is not allowed:

            >>> assert_valid_channel_name('channel 2')
            Traceback (most recent call last):
            ...
            InvalidChannelName: Special characters are not allowed in channels, first bad character is " "

        #.  When bad character matches at start of line, it's found:

            >>> assert_valid_channel_name('$1 channel')
            Traceback (most recent call last):
            ...
            InvalidChannelName: Special characters are not allowed in channels, first bad character is "$"

        #.  When name is just bad characters, first bad char is returned:

            >>> assert_valid_channel_name('#&()`')
            Traceback (most recent call last):
            ...
            InvalidChannelName: Special characters are not allowed in channels, first bad character is "#"

        #.  Some special characters are OK:

            >>> assert_valid_channel_name('-ABC_project=+/.,@1234')
    """
    if not name:
        raise InvalidChannelName('Empty channel name is not allowed')

    valid_channel_name = re.match(r'^[A-Za-z0-9=\+\-/\.,_@]*', name).group()

    if valid_channel_name == name:
        return

    raise InvalidChannelName(
        'Special characters are not allowed in channels, first bad character is "{}"'.format(
            name[len(valid_channel_name)]
        )
    )



from __future__ import absolute_import, print_function, unicode_literals

from .admin_client import AdminClient
from .database import Database
from .document import Document
from .query import Query
from .session import Session
from .stats import Stats
from .user import User
from .user_client import UserClient

__all_ = [
    AdminClient,
    Database,
    Document,
    Query,
    Session,
    Stats,
    User,
    UserClient,
]



from __future__ import absolute_import, print_function, unicode_literals

from .helpers import assert_valid_channel_name, assert_valid_document_id
from .resource import Resource


class Document(Resource):
    """
    A Couchbase Document in a database.

    Attributes:
        channels (tuple (str)): Channels this document is in.
        data (DataDict): Data from the Document using the `DataDict` manager.
            The `DataDict` instance prevents protected keys from entering the
            data, but does nothing to prevent mutation. Therefore it never
            contains the private SG fields '_id', '_rev', 'channels'.
        doc_id (str): ID of document.
        rev (str): Revision identifier of document. Set to empty string when no
            document has been retrieved.
        url (str): URL for this resource on Sync Gateway.
    """

    def __init__(self, database, doc_id):
        """
        Args:
            database (Database)
            doc_id (str)

        Raises:
            InvalidDocumentID: When doc_id is not valid.
        """
        super(Document, self).__init__(database)
        assert_valid_document_id(doc_id)
        self.doc_id = doc_id
        self.rev = ''
        self.channels = ()
        self.url = '{}{}'.format(self.database.url, self.doc_id)

    def set_channels(self, *channels):
        """
        Validate each channel passed and save to channels attribute. No update
        to channels is made if one is bad, all are rejected.

        Args:
            *args (str): Channels to be set for this document.

        Returns:
            None

        Raises:
            InvalidChannelName: When a bad channel name is passed.
        """
        for channel in channels:
            assert_valid_channel_name(channel)
        self.channels = channels

    def set_rev(self, revision_id):
        """
        Set Document's revision id. This is used for subsequent updates and
        deletions. Not checked for any validity.

        Args:
            revision_id (str)

        Returns:
            None

        Raises:
            ValueError: When revision_id is empty string.
        """
        if not revision_id:
            raise ValueError('Empty revision received')
        self.rev = revision_id

    def create_update(self):
        """
        Save or update Document in Sync Gateway. Saves the received revision id
        into instance's ``rev`` attribute.

        ``PUT /<database_name>/<doc_id>``

        Note:
            Works for updates but is not tested.

        Returns:
            int: ``AdminClient.CREATED`` if document was created (matches 201).
        """

        put_data = self.data.to_dict()

        if self.channels is not None:
            put_data['channels'] = self.channels
        if self.rev is not None:
            put_data['_rev'] = self.rev

        response = self.database.client.put(self.url, data=put_data)

        if response.status_code == 201:
            response_data = response.json()
            self.set_rev(response_data['rev'])
            return self.database.client.CREATED

        return False

    def retrieve(self):
        """
        Load document contents. Once loaded, `_rev` and `channels` are used to
        update the internal attributes before the data is sent to the DataDict.

        ``GET /<name>/<doc_id>``

        Returns:
            bool: Load was successful.

        Raises:
            DoesNotExist: Document with provided doc_id can not be loaded.

        Note:
            DataDict never contains the private Sync Gateway fields ``_id``,
            ``_rev``, ``channels``.

        Note:
            **Not** running through ``/<name>/_raw/<doc_id>``.

        Warning:
            Side effect: ``self.channels`` is updated based on returned the
            JSON using ``self.set_channels()``.

        Warning:
            Side effect: ``self.data`` is updated with data dictionary loaded
            from JSON.

        Warning:
            Side effect: ``self.rev`` is updated from revision passed in JSON
            using ``self.set_rev``.
        """
        response = self.database.client.get(self.url)
        response_data = response.json()

        self.set_rev(response_data['_rev'])
        if 'channels' in response_data:
            self.set_channels(*response_data['channels'])

        self.data = response_data

        return True

    def delete(self):
        """
        Delete Document from its Database. Document must have been retrieved in
        order for a valid revision ID to be provided. If there isn't a cache of
        this information when a `delete` is asked for, then a pre-fetch will
        occur.

        Uses the default ``Resource.delete`` action, but then inspects the
        response to ensure that ``ok`` is ``True``.

        ``DELETE /<name>/<doc_id>?rev=<rev>``

        Returns:
            bool: Delete was successful.

        Raises:
            DoesNotExist: If Document can't be found (doc has to be loaded
                first to retrieve the revision number, which can yield a 404 if
                it doesn't exist). Also can be raised if the Database does not
                exist.
            RevisionMismatch: If provided `rev` parameter does not match the
                live revision ID at request time.
        """
        if not self.rev:
            self.retrieve()

        params = {
            'rev': self.rev,
        }

        response = self.database.client.delete(self.url, params=params)

        return response.status_code == 200 and response.json()['ok']



from __future__ import absolute_import, print_function, unicode_literals

from .data_dict import DataDict
from .helpers import ComparableMixin


class Resource(ComparableMixin, object):
    """
    A Couchbase object stored within a Database, identified by a URL and
    accessible through REST verbs. Data is stored in a DataDict manager.

    Attributes:
        _data (DataDict)
        database (Database)
    """

    def __init__(self, database):
        """
        Initialise a Resource with a database.

        Args:
            database (Database)
        """
        if not getattr(database, 'url', None):
            raise ValueError(
                '{class_name} needs a `database` that provides a populated '
                '`url` (usually a `Database` instance), not {found}'.format(
                    class_name=self.__class__.__name__,
                    found=type(database).__name__,
                ),
            )
        self.database = database
        self.data = {}

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self._data = DataDict.from_dict(data)

    def __lt__(self, other):
        return self.url < other.url

    def __hash__(self):
        return hash(self.url)

    def __repr__(self):
        """
        Provides 'nice' output if object has defined `url` attribute. Else
        falls back to default.
        """
        try:
            nice_output = unicode(self)
            return nice_output.encode('UTF8')
        except (AttributeError, UnicodeEncodeError):
            return super(Resource, self).__repr__()

    def __unicode__(self):
        return '<{class_name} "{url}">'.format(
            class_name=self.__class__.__name__,
            url=self.url,
        )



from __future__ import absolute_import, print_function, unicode_literals

import six

from .exceptions import DoesNotExist, InvalidPassword
from .helpers import assert_valid_channel_name
from .resource import Resource


class User(Resource):
    """
    A User in a Database.

    Attributes:
        admin_channels (tuple(str)): Channels that this User has been added to
            by admin or sync function.
        data (DataDict)
        name (str): Username of this user.
        password (str): User's password. Can be `None` and is only set with
            `set_password`.
        retrieved (bool): User been retrieved from sync gateway. Acts as a flag
            to know if there should be a password sent at `create_update` time.
        url (str): URL for this User on sync gateway.
    """

    def __init__(self, database, name):
        """
        Args:
            database (Database)
            name (str): User name.
        """
        super(User, self).__init__(database)
        self.admin_channels = ()
        self.name = name
        self.password = None
        self.retrieved = False
        self.url = '{}_user/{}'.format(self.database.url, self.name)

    def set_password(self, password):
        """
        Args:
            password (str): A password for the User. Must not be empty (this is
                a basic security measure and is not enforced by sync gateway.

        Raises:
            InvalidPassword: Password provided for User was invalid.
        """
        if not password or not isinstance(password, six.string_types):
            raise InvalidPassword('"{}" is not a valid password for a User'.format(password))
        self.password = password

    def set_admin_channels(self, *channels):
        """
        Validate each channel passed in `channels` and save to `admin_channels`
        attribute. No update to channels is made if one is bad, all are
        rejected.

        Args:
            *args: Variable number of channels as str.

        Raises:
            InvalidChannelName: Bad channel name is received.
        """
        for channel in channels:
            assert_valid_channel_name(channel)
        self.admin_channels = channels

    def create_update(self):
        """
        Create a new or update an existing user for currently connected
        database. When creating a new User then their password must have been
        set on this instance using `set_password`, or the sync gateway will
        reject the PUT with a 400.

        NOTE it's not possible to change the username (name) of a user with
        this function.

        NOTE Does not handle roles, email or disabled state.

        `PUT /<database_name>/_user/<name>`

        Returns:
            int: `AdminClient.CREATED` if a new user was created (matches 201),
                `AdminClient.UPDATED` if an existing user was updated,

        Raises:
            DoesNotExist: When Database does not exist.
            SyncGatewayClientErrorResponse: When sync gateway returns a client
                error HTTP code.
        """
        data = {
            'admin_channels': self.admin_channels,
            'name': self.name,
        }

        if not self.retrieved:
            data['password'] = self.password

        response = self.database.client.put(self.url, data)

        if response.status_code == 201:
            return self.database.client.CREATED
        elif response.status_code == 200:
            return self.database.client.UPDATED

        # If code reaches here, there's an unexpected HTTP status code.

    def retrieve(self):
        """
        Get User's info from sync gateway.

        When User is found the response's payload is kept in the `data`
        attribute.

        When User is not found (404), any existing `data` is erased.
        Sync gateway response can optionally contain an `admin_channels` field.
        When none are returned then `admin_channels` attribute of the User is
        set to the empty tuple.

        `GET /<database_name>/_user/<name>`

        Returns:
            bool: Retrieval was successful.

        Raises:
            DoesNotExist: User can't be found in this Database or Database does
                not exist.
        """
        try:
            response = self.database.client.get(self.url)
        except DoesNotExist:
            self.data = {}
            raise

        self.data = response.json()
        self.retrieved = True

        try:
            self.set_admin_channels(*self.data['admin_channels'])
        except KeyError:
            self.set_admin_channels()

        return True

    def delete(self):
        """
        Delete User from Database.

        Returns:
            bool: User was deleted.

        Raises:
            DoesNotExist: User can't be found in this Database or Database does
                not exist.
        """
        response = self.database.client.delete(self.url)
        result = response.status_code == 200
        if result:
            self.retrieved = False
        return result



from __future__ import absolute_import, print_function, unicode_literals

from .client import Client


class AdminClient(Client):
    """
    Sync Gateway admin client for performing actions on the private admin API.

    Attributes:
        url (str): Sync Gateway admin REST API URL.
    """

    def __repr__(self):
        return '<AdminClient on "{url}">'.format(url=self.url)

    def __lt__(self, other):
        """
        Two AdminClient instances are the same if they are both talking to the
        same URL (there is no username / password to distinguish them).

        Returns:
            bool

        Raises:
            ValueError: When `other` is not an AdminClient.
        """
        if not isinstance(other, AdminClient):
            raise ValueError('AdminClient compared to {}'.format(type(other)))
        return self.url < other.url

    # --- Databases ---

    def all_databases(self):
        """
        Provide all Databases on the server.

        GET /_all_dbs

        Returns:
            list (Database): All databases found, connected with this client.

        Raises:
            GatewayDown: When sync gateway instance can not be reached by
                client.
        """
        response = self.get('{}{}'.format(self.url, '_all_dbs')).json()
        return [self.get_database(name) for name in response]



from __future__ import absolute_import, print_function, unicode_literals

from .document import Document
from .exceptions import DoesNotExist
from .helpers import ComparableMixin, assert_valid_database_name
from .query import Query
from .session import Session
from .user import User


class Database(ComparableMixin, object):
    """
    A Database on Sync Gateway.

    Attributes:
        client (AdminClient)
        name (str)
        url (str): URL to the database, created at init time, including
            trailing slash.
    """

    def __init__(self, client, name):
        """
        Args:
            client (AdminClient)
            name (str): A valid database name.
        """
        if not getattr(client, 'url', None):
            raise ValueError(
                '{class_name} needs a `client` that provides a populated '
                '`url` (usually a `AdminClient` instance), not {found}'.format(
                    class_name=self.__class__.__name__,
                    found=type(client).__name__,
                ),
            )

        assert_valid_database_name(name)

        self.client = client
        self.name = name
        self.url = '{}{}/'.format(self.client.url, self.name)

    def __repr__(self):
        return '<Database "{url}">'.format(url=self.url)

    def __lt__(self, other):
        """
        Comparison is only carried out on the url, however this will call the
        current client's settings and use them to build the url each time.

        Args:
            other (Database)

        Returns:
            bool

        Raises:
            ValueError: When other is not Database.
        """
        if not isinstance(other, Database):
            raise ValueError('Database compared to {}'.format(type(other)))
        return self.url < other.url

    def create(self):
        """
        Write this Database instance to Sync Gateway.

        Uses test orientated settings (i.e. none - the empty dictionary ``{}``
        is passed as data) to create database. This function is intended for
        test functionality, rather than for clients to be regularly creating
        databases.

        ``PUT /:name/``

        Returns:
            bool: Creation was successful.
        """
        return self.client.put(self.url, {}).status_code == 201

    def get(self):
        """
        Return information about this Database from Sync Gateway.

        ``GET /:name/``

        Returns:
            dict: Information loaded from SG.

        Raises:
            DoesNotExist: When database is not written to Sync Gateway
                regardless of whether the client is authorized or not.
            ClientUnauthorized: When database exists and client is not
                authorized.
        """
        response = self.client.get(self.url)
        return response.json()

    def delete(self):
        """
        Remove database.

        Whereas Sync Gateway will raise 404 if the database is not found, this
        fails silently with the intention that it can be used 'scatter gun'
        style at the end of test runs to clean up database lists.

        ``DELETE /:name/``

        Returns:
            bool: Database was found and deleted.
        """
        try:
            response = self.client.delete(self.url)
        except DoesNotExist:
            return False
        return response.status_code == 200

    # --- Documents ---

    def get_document(self, doc_id):
        """
        Returns:
            Document: An instance of Document in this Database with provided
            ``doc_id``.
        """
        return Document(self, doc_id)

    def all_docs(self):
        """
        Get list of all Documents in database.

        ``GET /:name/_all_docs``

        NOTE Use for testing only. From Simon @ Couchbase:

            We would strongly advise against using the `_all_docs` endpoint. As
            your database grows relying on the View that this calls to return
            to you every document key is inadvisable and does not scale well to
            very high numbers of documents.

            If you need to retrieve or update multiple documents please use the
            _bulk_get and _bulk_docs end points to supply a list of keys (or
            documents) for retrieval or update.

        Returns:
            list (Document): An instance of Document for each document returned
            by the endpoint. For each instance the ``data['_rev']`` value is
            populated with the revision ID from ``value.rev``.

        Raises:
            DoesNotExist: Database can't be found on sync gateway.
        """
        url = '{}{}'.format(self.url, '_all_docs')

        response = self.client.get(url)

        documents = []

        for doc_info in response.json()['rows']:
            document = self.get_document(doc_info['id'])
            document.set_rev(doc_info['value']['rev'])
            documents.append(document)

        return documents

    # --- Users ---

    def get_user(self, username):
        """
        Returns:
            User: An instance of User for the provided ``username``.
        """
        return User(self, username)

    def all_users(self):
        """
        ``GET /:name/_user/``

        Returns:
            list (User): All Users in Database.
        """
        url = '{}_user/'.format(self.url)
        response = self.client.get(url)
        return [self.get_user(username) for username in response.json()]

    # --- Queries ---

    def get_query(self, doc_id):
        """
        Returns:
            Query: An instance of a query design document in this Database with
            the provided ``doc_id``.
        """
        return Query(self, doc_id)

    # --- Session ---

    def get_session(self):
        """
        Returns:
            Session
        """
        return Session(self)



from __future__ import absolute_import, print_function, unicode_literals

import json

from .resource import Resource


class Query(Resource):
    """
    Query a design document.

    Attributes:
        data (DataDict): Data from the design document using the ``DataDict``
            manager.
        doc_id (str): ID of design document.
        url (str): URL for this resource on Sync Gateway.
    """

    def __init__(self, database, doc_id):
        """
        Args:
            database (Database)
            doc_id (str): Design document name.
        """
        super(Query, self).__init__(database)
        self.doc_id = doc_id
        self.url = '{}_design/{}'.format(self.database.url, self.doc_id)

    def build_view_url(self, view_name):
        """
        Args:
            view_name (str)

        Returns:
            str: URL for querying view.
        """
        return '{}/_view/{}'.format(self.url, view_name)

    def create_update(self):
        """
        Create or update design document with data.

        ``PUT /<database_name>/_design/<doc_id>``

        Returns:
            bool: Design document was created or updated successfully.
        """
        return self.database.client.put(self.url, data=self.data).status_code == 201

    def retrieve(self):
        """
        Returns:
            bool: Design document was retrieved.

        Raises:
            DoesNotExist: Design document or Database can not be found.

        Side effects:
            data: Updates internal data dictionary with data loaded from JSON.
        """
        response = self.database.client.get(self.url)
        self.data = response.json()
        return True

    def delete(self):
        """
        Delete design document.

        Returns:
            bool: Design document deleted.

        Raises:
            DoesNotExist: Design document or Database can not be found.
        """
        return self.database.client.delete(self.url).status_code == 200

    def query_view(self, view_name, key=None, stale=True, timeout=None):
        """
        Load a view function from this design Document. Document must be
        written to Sync Gateway and view's Javascript function must be valid.

        Args:
            view_name (str): View's name.
            key (Optional): Value to use to search the view's key (the
                left part of the map). Any type can be passed as long as:

                * ``key is not None``

                * ``key`` can be serialized to a string by ``json.dumps()``.

                To query a view with multiple keys set ``key`` as an iterable
                which will be serialized to JSON as an array. E.g.
                ``key=['left_id', 'right_id']``.
            stale (bool, Optional): Allow stale results in the view. This is
                currently the default value in Sync Gateway, so is only passed
                when set to ``False``. Default ``True``.

                ``'false'`` is an undocumented option for this param. See
                https://github.com/couchbase/sync_gateway/issues/727#issuecomment-83588984

                It also doesn't work in Walrus mode, see
                https://github.com/constructpm/pysyncgateway/issues/7
            timeout (int, Optional): Set a time out of seconds as per requests'
                spec which means that if the Sync Gateway does not respond to
                the GET request within the ``timeout`` period, a ``ReadTimeout``
                will be raised.  Default ``None`` which means that requests'
                default is used.

        Returns:
            dict: Decoded JSON for view data result. Will usually be a
            dictionary contain keys 'Collator', 'rows' and 'total_rows'.

        Raises:
            DoesNotExist: When database, design document or contained view can
                not be found.
            requests.exceptions.ReadTimeout: When Sync Gateway does not respond
                within given ``timeout``.
        """
        # Build params (passed to SG in the URL)
        params = {}
        if not stale:
            params['stale'] = 'false'
        if key is not None:
            params['key'] = json.dumps(key)

        # Build kwargs (passed to requests.get)
        kwargs = {}
        if params:
            kwargs['params'] = params
        if timeout is not None:
            kwargs['timeout'] = timeout

        url = self.build_view_url(view_name)
        response = self.database.client.get(url, **kwargs)

        return response.json()



from __future__ import absolute_import, print_function, unicode_literals

import six

from .exceptions import InvalidDataKey


class DataDict(dict):
    """
    DataDict is the developer facing dictionary of data contained within a
    Couchbase Document. It prevents settings items with reserved Couchbase keys
    like "_rev" or "_id", but still acts like a dictionary to make manipulating
    Document data easy.

    NOTE: It is not design agnostic because it protects the "channels" key.
    Depending on your application and sync function, that might not be a list
    of channels in your data design.
    """

    filtered_keys = (
        '_id',
        '_rev',
        'channels',
    )

    @classmethod
    def from_dict(obj, data):
        """
        Given a dictionary `data`, create a new DataDict from a dictionary
        `data`, silently removing all filtered keys from that input.

        Args:
            data (dict): Input data.

        Returns:
            DataDict: New instance created with cleaned, copied `data`.

        Raises:
            ValueError: When passed a non-dict.
        """
        if not isinstance(data, dict):
            raise ValueError('data argument is not dict')
        new = obj()
        for key, value in six.iteritems(data):
            try:
                new[key] = value
            except InvalidDataKey:
                pass
        return new

    def __setitem__(self, key, value):
        """
        Assign value to key as usual, but only if `key` is not in
        `filtered_keys` list.

        Args:
            key (str)
            value

        Raises:
            InvalidDataKey: When data is passed that is not managed by
                DataDict.
        """
        if key in self.filtered_keys:
            raise InvalidDataKey('DataDict does not allow "{}" key'.format(key))
        return super(DataDict, self).__setitem__(key, value)

    def to_dict(self):
        """
        Returns:
            dict: A dictionary version of self
        """
        return dict(self)



from __future__ import absolute_import, print_function, unicode_literals

__name__ = 'pysyncgateway'
__version__ = '1.1.0'

__author__ = 'Construct Technology Ltd'
__copyright__ = '2018, {}'.format(__author__)

__description__ = 'Library for communication with Couchbase Sync Gateway'
__email__ = 'info@construct.pm'



from __future__ import absolute_import, print_function, unicode_literals

from .exceptions import NotLoaded
from .resource import Resource


class Session(Resource):
    """
    Sessions on the Database are undocumented as of Sync Gateway 1.5 in both
    the `public
    <https://developer.couchbase.com/documentation/mobile/1.5/references/sync-gateway/rest-api/index.html#/>`_
    and `admin
    <https://developer.couchbase.com/documentation/mobile/1.5/references/sync-gateway/admin-rest-api/index.html>`_
    REST APIs. But since they provide a useful test mechanism by showing a list
    of channels that the authenticated User has been subscribed to, they are
    included in this library.

    Attributes:
        data (DataDict): Data from the session using the ``DataDict`` manager.
        database (Database)
        url (str): URL for the session in the Database on Sync Gateway.
    """

    def __init__(self, database):
        """
        Args:
            database (Database)
        """
        super(Session, self).__init__(database)
        self.url = '{}_session'.format(self.database.url)

    def retrieve(self):
        """
        Collect session information for the current client in the ``database``
        attribute.

        For unauthorized users on the public API and requests on the admin API
        this looks like::

            {
                'authentication_handlers': ['default', 'cookie'],
                'ok': True,
                'userCtx': {
                    'channels': {},
                    'name': None,
                },
            }

        For authenticated users on the public API, in this example with the
        name ``__USERNAME__`` and channels ``a``, ``b`` and ``c``, this looks
        like::

            {
                'authentication_handlers': ['default', 'cookie'],
                'ok': True,
                'userCtx': {
                    'channels': {
                        '!': 1,
                        'a': 1,
                        'b': 1,
                        'c': 1,
                    },
                    'name': '__USERNAME__',
                },
            }

        Returns:
            bool: Success

        Raises:
            DoesNotExist: When database does not exist on Sync Gateway,
                regardless of whether client is authorized or not.

        Warning:
            Side effect: Updates ``self.data`` with response from Sync Gateway
            on success.
        """
        response = self.database.client.get(self.url)
        response_data = response.json()
        self.data = response_data
        return True

    def _get_user_ctx(self):
        try:
            return self.data['userCtx']
        except KeyError:
            raise NotLoaded()

    def get_name(self):
        """
        Helper to pick user's ``name`` field from retrieved session data.

        Returns:
            str: Name of user according to Sync Gateway.

        Raises:
            NotLoaded: When session has not been retrieved.
            KeyError: When the ``userCtx`` data in the session response does
                not contain a ``name`` field.
        """
        return self._get_user_ctx()['name']

    def get_channels(self):
        """
        Helper to pick user's subscribed channels from retrieved session data.

        Returns:
            list: Sorted list of channel names found in response including the
            special ``!`` public channel.

        Raises:
            NotLoaded: When session has not been retrieved.
            KeyError: When the ``userCtx`` data in the session response does
                not contain a ``channels`` dictionary.
        """
        return sorted(list(self._get_user_ctx()['channels']))



from __future__ import absolute_import, print_function, unicode_literals

import os

from setuptools import setup

basedir = os.path.dirname(__file__)


def readme():
    with open(os.path.join(basedir, 'README.rst')) as f:
        return f.read()


about = {}
with open(os.path.join(basedir, 'pysyncgateway', '__about__.py')) as f:
    exec (f.read(), about)

setup(
    name=about['__name__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme(),
    url='https://github.com/constructpm/pysyncgateway',
    author=about['__author__'],
    author_email=about['__email__'],
    license='Apache License 2.0',
    install_requires=[
        'requests>=2.18',
        'six>=1.10',
    ],
    packages=['pysyncgateway'],
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python',
    ],
)



from __future__ import absolute_import, print_function, unicode_literals

from requests import delete, get, put

from .database import Database
from .helpers import ComparableMixin, sg_method


class Client(ComparableMixin, object):
    """
    Abstract parent class for `AdminClient` and `UserClient`.

    Attributes:
        _auth (requests.HTTPBasicAuth): Initialises to `None` and is only used
            by `UserClient`.
        url (str): Sync Gateway REST API URL.
    """
    CREATED = 1
    UPDATED = 2
    CONFLICT = 3

    def __init__(self, url):
        """
        Args:
            url (str): Sync Gateway REST API URL (excluding database but
                including trailing slash).
        """
        self.url = url
        self._auth = None

    def get_server(self):
        """
        Returns:
            dict: Meta-information about the server.
        """
        return self.get(self.url).json()

    def get_database(self, database_name):
        """
        Get a `Database` instance connected to this client.

        Args:
            database_name (str): Name of database.

        Returns:
            Database
        """
        return Database(self, database_name)

    # --- REST Verbs ---

    @sg_method
    def get(self, url, **kwargs):
        if self._auth:
            kwargs['auth'] = self._auth
        return get(url, **kwargs)

    @sg_method
    def put(self, url, data):
        return put(url, json=data)

    @sg_method
    def delete(self, url, **kwargs):
        return delete(url, **kwargs)



from __future__ import absolute_import, print_function, unicode_literals

from requests.auth import HTTPBasicAuth

from .client import Client


class UserClient(Client):
    def auth(self, un, pw):
        """
        Authorise client with provided credentials. Does not check with Sync
        Gateway that credentials are correct until a request is made.

        Args:
            un (str): User name.
            pw (str): Password.

        Returns:
            None
        """
        self._auth = HTTPBasicAuth(un, pw)



from __future__ import absolute_import, print_function, unicode_literals


class PysyncgatewayException(Exception):
    """
    The root of all Evil >:D
    """


class ClientUnauthorized(PysyncgatewayException):
    """
    Client is not authorized to access this URL
    """


class GatewayDown(PysyncgatewayException):
    """
    SyncGateway could not be reached on configured URL
    """


class InvalidChannelName(PysyncgatewayException):
    pass


class InvalidDataKey(PysyncgatewayException):
    pass


class InvalidDatabaseName(PysyncgatewayException):
    pass


class InvalidDocumentID(PysyncgatewayException):
    pass


class InvalidPassword(PysyncgatewayException):
    pass


class DoesNotExist(PysyncgatewayException):
    """
    Generic exception to replace 404s. Used if databases, users or documents
    can't be loaded.
    """


class NotLoaded(PysyncgatewayException):
    """
    Item from a ``Resources`` data attribute was requested, but that resource
    has not yet been successfully retrieved from Sync Gateway.
    """


class RevisionMismatch(PysyncgatewayException):
    pass


class SyncGatewayClientErrorResponse(PysyncgatewayException):
    """
    Sync gateway responded with a 4xx error.

    Attributes:
        status_code (int): Error code in the response from sync gateway.
        json (dict): Body of response from sync gateway.
    """

    def __init__(self, status_code, json):
        self.status_code = status_code
        self.json = json

    @classmethod
    def from_response(obj, response):
        """
        Args:
            response (requests.Response)

        Returns:
            SyncGatewayClientErrorResponse
        """
        return obj(response.status_code, response.json())

    def __repr__(self):
        """
        Returns:
            str
        """
        return '<SyncGatewayClientErrorResponse {}>'.format(self)

    def __str__(self):
        """
        Used by pytest to show stack trace.

        Returns:
            str: Containing `status_code` and `json['error']`.
        """
        return '{} "{}"'.format(self.status_code, self.json['error'])



from __future__ import absolute_import, print_function, unicode_literals


class Stats(object):
    """
    Stats object from the expvars endpoint on the Sync Gateway. See
    https://github.com/couchbase/sync_gateway/wiki/expvars

    NOTE: Stats come from the server and not from the database.

    Attributes:
        client (AdminClient): Used to communicate with the server.
        data (dict): Statistical data populated after retrieval. Will be empty
            dictionary before retrieval.
        url (str): Location of stats.
    """

    def __init__(self, client):
        """
        Args:
            client (AdminClient): Used to communicate with the server.
        """
        self.client = client
        self.url = '{}_expvar'.format(self.client.url)
        self.data = {}

    def retrieve(self):
        """
        Load stats, parse info and stash in data attr.

        Returns:
            bool: load was successful.

        Raises:
            GatewayDown: When the endpoint can't be reached.
            ValueError: When something non-JSON is loaded.

        Side effects:
            data: Populated with dictionary parsed from the JSON response.
        """
        response = self.client.get(self.url)
        self.data = response.json()
        return True
